module Trie where

import List
import Maybe
data (Ord a) => TrieBr a =   I (a, Trie a)
                            | E (a, [a])
                    deriving (Read,Show)

newtype (Ord a) => Trie a = Trie (Int, [TrieBr a])
                    deriving (Read,Show)

data TrieLvlFound a =   LvlFound (TrieBr a)
                      | LvlNotFound

triefind:: (Ord a)=>[a]->Trie a->Bool

triefind xs t =
   let
     nd = (trieconflictsearch xs 0 t)
   in if null nd
      then False
      else isPrefixOf xs (leastLeaf nd)

leastLeaf (E(_,x):_) = x
leastLeaf (I(_,Trie (_,(x:xs))):_) = leastLeaf (x:xs)

trieconflictsearch:: (Ord a)=> [a]->Int->Trie a->[TrieBr a]
trieconflictsearch [] _  (Trie (p, t)) = t
trieconflictsearch xs ndropped (Trie (p, t)) =
   let 
     suffxs = drop (p-ndropped) xs
   in if null suffxs
      then t
      else case trienodefind (head suffxs) t of
        	(LvlFound (I (y,yt))) -> trieconflictsearch suffxs p yt
        	(LvlFound (E l))    -> [(E l)]
        	LvlNotFound    -> t



trienodefind:: (Ord a)=> a->[TrieBr a]->TrieLvlFound a
trienodefind x [] = LvlNotFound
trienodefind x (I (y, yt):ts) =  
   if x == y
   then LvlFound (I(y,yt))
   else trienodefind x ts
trienodefind x (E (y, s):ts) = 
   if x == y
   then LvlFound (E(y,s))
   else trienodefind x ts

newleaf:: (Ord a)=> a->[a]->TrieBr a
newleaf x xs = E(x,xs)


newtrie::(Ord a) => [a]->[a]->Trie a
newtrie xs ys =  Trie(pos,insertBy cmpBrBr (E(xsch,xs)) [E(ysch,ys)])
    where (pos,xsch,ysch) = firstDiff 0 xs ys
   
trieinsert:: (Ord a) => [a]->Trie a->Trie a
trieinsert xs t =
   let 
     d = triedisagreement xs t
   in
     if  isNothing d
     then t
     else trienodeinsert (fromJust d)  t


triedisagreement:: (Ord a)=>[a]->Trie a->Maybe (Int,a,[a],a,[a])
triedisagreement xs t = 
      if isPrefixOf xs (leastLeaf conflict)
      then Nothing
      else Just (disagreement xs (leastLeaf conflict))
         where conflict = trieconflictsearch xs 0 t

disagreement xs ys = (diffpos,xsch,xs,ysch,ys)
   where (diffpos,xsch,ysch) = firstDiff 0 xs ys

firstDiff pos (x:xs) (y:ys) =
            if x==y
            then firstDiff (pos+1) xs ys
            else (pos,x,y)

trienodeinsert (disagg@(diffpos,xinch,xin,ysch,ys)) (Trie(d,brlst)) = 
    if diffpos < d 
    then Trie(diffpos, insertBy cmpBrBr (E(xinch,xin)) [I(ysch,Trie(d,brlst))])
    else if diffpos == d
    then Trie(d,insertBy cmpBrBr (E(xinch,xin)) brlst)
    else Trie(d,initbrlst++(newbr:(tail restbrlst)))
            where restbrlst = dropWhile (cmpChBr (xin!!d)) brlst
                  initbrlst = takeWhile (cmpChBr (xin!!d)) brlst
                  newbr = branchinsert disagg (head restbrlst)

 
branchinsert disagg (I(c,t)) = I(c,trienodeinsert disagg t)
branchinsert (_,_,xin,_,ys) (E(c,_)) = I(c,newtrie xin ys)
   
cmp c d = 
   if c<d
   then LT
   else if c==d
   then EQ
   else GT

cmpChBr c (E(d,_)) = c > d
cmpChBr c (I(d,_)) = c > d
cmpBrBr (E(c,_)) (E(d,_)) = cmp c d
cmpBrBr (E(c,_)) (I(d,_)) = cmp c d
cmpBrBr (I(c,_)) (E(d,_)) = cmp c d
cmpBrBr (I(c,_)) (I(d,_)) = cmp c d
         
 
s2 = "g"
s1 = "cg"
s0 = "ccg"
n1 = newtrie s0 s1
t = trieinsert s2 n1


